# This file is @generated by <https://github.com/liblaf/copier-python>.
# DO NOT EDIT!

# ref: <https://github.com/ofek/hatch-vcs>
__version__: str
__version_tuple__: tuple[int | str, ...]
version_tuple: tuple[int | str, ...]
version: str
